==> ben-2.16.tar.gz: ben/snapshot.c <==
/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "xmem.h"
#include "common.h"
#include "ini.h"
#include "snapshot.h"



// ***************************************************************************
// 
// ***************************************************************************
#define DEBUG(...)	\
	do { \
		fprintf(stderr, "%s ", curdate_str()); \
		fprintf(stderr, __VA_ARGS__); \
	} while(0)

// ***************************************************************************
// 
// ***************************************************************************
static void snap_job(buffer *b, nodeset *clients, node *c, job *t, int type)
{
	ini_section(b, "job");
	
	if (type == PROTO_TASK_DONE) {
		ini_entry(b, "type", "done");
	} else if (type == PROTO_TASK_RUNNING) {
		ini_entry(b, "type", "running");
	} else if (type == PROTO_TASK_PENDING) {
		ini_entry(b, "type", "pending");
	} else {
		ini_entry(b, "type", "unknown");
	}
	
	ini_int(b, "id", t->id);
	
	ini_entry(b, "shell", t->shell);
	ini_entry(b, "dir", t->dir);
	ini_entry(b, "options", t->options);
	ini_entry(b, "name", t->name);
	ini_entry(b, "command", t->cmd);
	
	if (c) {
		ini_int(b, "running_id", c->id);
		ini_entry(b, "running_name", c->name);
	}
	
	if (t->restrict_box >= 0) {
		ini_int(b, "restrict_id", t->restrict_box);
		
		node *fc = nodeset_find(clients, t->restrict_box);
		
		if (fc)
			ini_entry(b, "restrict_name", fc->name);
	}

	if (t->forward_box >= 0) {
		ini_int(b, "forward_id", t->forward_box);
		
		node *fc = nodeset_find(clients, t->forward_box);
		
		if (fc)
			ini_entry(b, "forward_name", fc->name);
	}
	
	if (t->ran_box >= 0) {
		ini_int(b, "ran_id", t->ran_box);
		
		node *fc = nodeset_find(clients, t->ran_box);
		
		if (fc)
			ini_entry(b, "ran_name", fc->name);
	}
	
	ini_entry(b, "stdout_path", t->outpath);
	ini_entry(b, "stderr_path", t->errpath);
	
	if (t->start_time)
		ini_date(b, "start_time", t->start_time);
	if (t->stop_time)
		ini_date(b, "stop_time", t->stop_time);
}

// ***************************************************************************
// 
// ***************************************************************************
static int snap_build(buffer *b, nodeset *clients, queue *done, queue *pending,
	char *path, char *tmp_path)
{
	int dcount = 0;
	
	for (job *t = done->head; t != NULL; t = t->next) {
		snap_job(b, clients, NULL, t, PROTO_TASK_DONE);
		dcount++;
	}
	
	int rcount = 0;
	for (node *c = clients->head; c != NULL; c = c->next) {
		for (job *t = c->running.head; t != NULL; t = t->next) {
			snap_job(b, clients, c, t, PROTO_TASK_RUNNING);
			rcount++;
		}
	}
	
	int pcount = 0;
	for (job *t = pending->head; t != NULL; t = t->next) {
		snap_job(b, clients, NULL, t, PROTO_TASK_PENDING);
		pcount++;
	}

	if (buffer_save_path(b, tmp_path))
		return -1;
	
	if (rename(tmp_path, path)) {
		DEBUG("rename('%s', '%s'): %s\n",
			tmp_path, path, strerror(errno));
		return -1;
	}
	
	DEBUG("snapshot to '%s': done %d, running %d, pending %d\n",
		path, dcount, rcount, pcount);
	
	return 0;
}

// ***************************************************************************
// 
// ***************************************************************************
int server_snapshot(nodeset *clients, queue *done, queue *pending,
	char *path)
{
	char *tmp_path = xprintf(NULL, "%s-tmp", path);
	buffer b;
	
	buffer_init(&b, 0);
	
	int r = snap_build(&b, clients, done, pending, path, tmp_path);
	
	buffer_clear(&b);
	
	xfree(tmp_path);
	
	return r;
}


// ***************************************************************************
// 
// ***************************************************************************

==> ben-2.17.tar.gz: ben/snapshot.c <==
/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "xmem.h"
#include "common.h"
#include "ini.h"
#include "snapshot.h"



// ***************************************************************************
// 
// ***************************************************************************
#define DEBUG(...)	\
	do { \
		fprintf(stderr, "%s ", \
			curdate_str(GTIME_DAYS | GTIME_SUBSEC).s); \
		fprintf(stderr, __VA_ARGS__); \
	} while(0)

// ***************************************************************************
// 
// ***************************************************************************
static void snap_job(buffer *b, nodeset *clients, node *c, job *t, int type)
{
	ini_section(b, "job");
	
	if (type == PROTO_TASK_DONE) {
		ini_entry(b, "type", "done");
	} else if (type == PROTO_TASK_RUNNING) {
		ini_entry(b, "type", "running");
	} else if (type == PROTO_TASK_PENDING) {
		ini_entry(b, "type", "pending");
	} else {
		ini_entry(b, "type", "unknown");
	}
	
	ini_int(b, "id", t->id);
	
	ini_entry(b, "shell", t->shell);
	ini_entry(b, "dir", t->dir);
	ini_entry(b, "options", t->options);
	ini_entry(b, "name", t->name);
	ini_entry(b, "command", t->cmd);
	
	if (c) {
		ini_int(b, "running_id", c->id);
		ini_entry(b, "running_name", c->name);
	}
	
	if (t->restrict_box >= 0) {
		ini_int(b, "restrict_id", t->restrict_box);
		
		node *fc = nodeset_find(clients, t->restrict_box);
		
		if (fc)
			ini_entry(b, "restrict_name", fc->name);
	}

	if (t->forward_box >= 0) {
		ini_int(b, "forward_id", t->forward_box);
		
		node *fc = nodeset_find(clients, t->forward_box);
		
		if (fc)
			ini_entry(b, "forward_name", fc->name);
	}
	
	if (t->ran_box >= 0) {
		ini_int(b, "ran_id", t->ran_box);
		
		node *fc = nodeset_find(clients, t->ran_box);
		
		if (fc)
			ini_entry(b, "ran_name", fc->name);
	}
	
	ini_entry(b, "stdout_path", t->outpath);
	ini_entry(b, "stderr_path", t->errpath);
	
	if (gtime_nonzero(t->wall_time_start))
		ini_date(b, "start_time", t->wall_time_start);
	if (gtime_nonzero(t->wall_time_stop))
		ini_date(b, "stop_time", t->wall_time_stop);
}

// ***************************************************************************
// 
// ***************************************************************************
static int snap_build(buffer *b, nodeset *clients, queue *done, queue *pending,
	char *path, char *tmp_path)
{
	gtime t0 = gtime_mono();
	
	int dcount = 0;
	
	for (job *t = done->head; t != NULL; t = t->next) {
		snap_job(b, clients, NULL, t, PROTO_TASK_DONE);
		dcount++;
	}
	
	int rcount = 0;
	for (node *c = clients->head; c != NULL; c = c->next) {
		for (job *t = c->running.head; t != NULL; t = t->next) {
			snap_job(b, clients, c, t, PROTO_TASK_RUNNING);
			rcount++;
		}
	}
	
	int pcount = 0;
	for (job *t = pending->head; t != NULL; t = t->next) {
		snap_job(b, clients, NULL, t, PROTO_TASK_PENDING);
		pcount++;
	}
	
	gtime t1 = gtime_mono();

	if (buffer_save_path(b, tmp_path))
		return -1;
	
	if (rename(tmp_path, path)) {
		DEBUG("rename('%s', '%s'): %s\n",
			tmp_path, path, strerror(errno));
		return -1;
	}
	
	gtime t2 = gtime_mono();
	
	size_t delta0 = gtime_get_ns(gtime_sub(t1, t0));
	size_t delta1 = gtime_get_ns(gtime_sub(t2, t1));
	size_t delta = gtime_get_ns(gtime_sub(t2, t0));
	size_t tr = b->size;
	
	DEBUG("snapshot '%s': d %d, r %d, p %d\n",
		path, dcount, rcount, pcount);
	DEBUG("serialization %zd ns, io %zd ns, %zd MB, %zd MB/s\n",
		delta0, delta1, tr >> 20, tr * 1000 / delta);
	
	return 0;
}

// ***************************************************************************
// 
// ***************************************************************************
int server_snapshot(nodeset *clients, queue *done, queue *pending,
	char *path)
{
	char *tmp_path = xprintf(NULL, "%s-tmp", path);
	buffer b;
	
	buffer_init(&b, 0);
	
	int r = snap_build(&b, clients, done, pending, path, tmp_path);
	
	buffer_clear(&b);
	
	xfree(tmp_path);
	
	return r;
}


// ***************************************************************************
// 
// ***************************************************************************
